indexing
	description: "A bitmap data structure of (more or less) arbitrary size.%
                %N.B. This is NOT a picture; It is a map of bits!!"
	license: "Eiffel Forumn License v2 - See license.txt"
	source: "{
		Amalasoft Corporation
		273 Harwood Avenue
		Littleton, MA 01460 USA
		}"
	status: "Copyright 1995-2008, Amalasoft";
	date: "$Date: 2008/12/03 $";
	revision: "$Revision: 001$";
	howto: "{
    To use this class, create a new instance with a given number of bits.
    All bits are unset on creation.
    Call set_position to set a given bit to non-zero and unset_poistion
    to clear that given bit.
    Call position_is_set with the desired bit position to see if that
    bit position is current set.
}"

class AEL_DS_BITMAP

inherit
	ANY
		redefine
			out
		end

create
	make

--|========================================================================
feature {NONE} -- Creation
--|========================================================================

	make (sz: INTEGER) is
			-- Create a new bitmap with 'sz' positions
			-- 'sz' MUST be an even multiple of 8
		require
			valid_size: sz > 0 and (sz \\8) = 0
		do
			count := sz
			create map.make_filled ('%U', length)
		end

--|========================================================================
feature -- Status
--|========================================================================

	map: STRING_8
			-- The actual map

	count: INTEGER
		-- Number of bits in map

	length: INTEGER is
		-- Number of BYTES in map
		do
			Result := count // 8
		end

	number_of_set_positions: INTEGER is
			-- Number of bit positions that are set
		do
			Result := bits_set.count
		end

--|========================================================================
feature -- Validation
--|========================================================================

	is_valid_index (v: INTEGER): BOOLEAN is
			-- Is the given number a valid index for this map?
		do
			Result := v > 0 and v <= count
		end

--|========================================================================
feature -- External representation
--|========================================================================

	out: STRING is
			-- String representation of bitmp (as a stream of 1s and 0s)
		local
			ti: NATURAL_8
			i, bi, lim: INTEGER
		do
			create Result.make (count)
			lim := length
			from i := 1
			until i > lim
			loop
				ti := map.item (i).code.to_natural_8
				from bi := 7
				until bi < 0
				loop
					if ti.bit_test (bi) then
						Result.extend ('1')
					else
						Result.extend ('0')
					end
					bi := bi - 1
				end
				i := i + 1
			end
		end

--|========================================================================
feature -- Access
--|========================================================================

	bits_set: LINKED_LIST [INTEGER] is
			-- List of bit positions that are set, starting at 1.
			-- See index_to_positions comment for more information
		local
			i, lim, bi: INTEGER
			byte: NATURAL_8
		do
			create Result.make
			lim := length
			from i := 1
			until i > lim
			loop
				byte := map.item (i).code.to_natural_8
				if byte /= 0x00 then
					-- Has a set bit
					from bi := 7
					until bi < 0
					loop
						if byte.bit_test (bi) then
							-- Found a set number
							Result.extend (positions_to_index (<< i, bi >>))
						end
						bi := bi - 1
					end
				end
				i := i + 1
			end
		end

	--|--------------------------------------------------------------

	first_unset_position: INTEGER is
			-- First available (unset) position
		local
			i, lim, pos, bi: INTEGER
			byte: NATURAL_8
		do
			lim := count
			from i := 1
			until i > lim or Result /= 0
			loop
				byte := map.item (i).code.to_natural_8
				if byte /= 0xff then
					-- Has a free bit, find the first one
					pos := ((i -1) * 8)
					from bi := 7
					until bi < 0 or Result /= 0
					loop
						if not byte.bit_test (bi) then
							-- Found a free number
							Result := pos + (bi - 7)
						else
							bi := bi - 1
						end
					end
				else
				end
				i := i + 1
			end
		end

	--|--------------------------------------------------------------

	highest_set_position: INTEGER is
			-- Highest position number set
		local
			i, lim, pos, bi: INTEGER
			byte: NATURAL_8
		do
			lim := count
			from i := lim
			until i = 0 or Result /= 0
			loop
				byte := map.item (i).code.to_natural_8
				if byte /= 0x00 then
					-- Has a set bit, find the highest one
					pos := ((i - 1) * 8)
					from bi := 0
					until bi >= 7 or Result /= 0
					loop
						if byte.bit_test (bi) then
							-- Found a set number
							Result := pos + (bi - 7)
						else
							bi := bi + 1
						end
					end
				end
				i := i - 1
			end
		end

	--|--------------------------------------------------------------

	position_is_set (v: INTEGER): BOOLEAN is
			-- is the given position set?
		local
			cpos, bpos: INTEGER
			positions: like index_to_positions
			byte: NATURAL_8
		do
			positions := index_to_positions (v)
			cpos := positions.item (1)
			bpos := positions.item (2)
			byte := map.item (cpos).code.to_natural_8
			Result := byte.bit_test (bpos)
		end

--|========================================================================
feature -- Status setting
--|========================================================================

	set_position (v: INTEGER) is
			-- Set the 'v-th' bit in the map
		require
			in_range: is_valid_index (v)
		do
			set_position_value (v, True)
		ensure
			is_set: position_is_set (v)
		end

	--|--------------------------------------------------------------

	unset_position (v: INTEGER) is
			-- Set the 'v-th' bit in the map
		require
			in_range: is_valid_index (v)
		do
			set_position_value (v, False)
		ensure
			is_set: not position_is_set (v)
		end

	--|--------------------------------------------------------------

	grow (v: INTEGER) is
			-- Increase the size of Current to 'v'
		require
			valid_size: v > 0 and (v \\8) = 0
			larger: v > count
		local
			ts: like map
		do
			create ts.make_filled ('%U', length - (v // 8))
			map.append (ts)
			count := v
		end

--|========================================================================
feature {NONE} -- Implementation
--|========================================================================

	index_to_positions (v: INTEGER): ARRAY [INTEGER] is
			-- The calculated byte and bit positions (in the map)
			-- corresponding to the given number.
			-- For example, '1' would be the first bit of the first
			-- byte in the map and would have a character (byte)
			-- position of 1, and a bit position of 0 (bit positions are
			-- right to left)
		require
			in_range: is_valid_index (v)
		local
			cpos, bpos: INTEGER
		do
			cpos := (v // 8)
			bpos := (v \\ 8)
			if bpos = 0 then
				-- Highest bit of cpos
				bpos := 7
			else
				-- (n - i)th bit of next byte, where 0 = LSB
				cpos := cpos + 1
				bpos := bpos - 1
			end
			Result := << cpos, bpos >>
		end

	--|--------------------------------------------------------------

	positions_to_index (v: ARRAY [INTEGER]): INTEGER is
			-- Index corresponding to the given byte and bit positions.
		require
			exists: v /= Void
			valid_positions: v.count = 2
			valid_byte: v.item (1) <= length
			valid_bit: v.item (2) >= 0 and v.item (2) <= 7
		local
			cval, bval: INTEGER
		do
			cval := (v.item (1) - 1) * 8
			bval := v.item (2) + 1
			Result := cval + bval
		ensure
			in_range: Result > 0 and Result <= count
		end

	--|--------------------------------------------------------------

	set_position_value (v: INTEGER; st: BOOLEAN) is
			-- Set the 'v-th' bit in the map
		require
			in_range: is_valid_index (v)
		local
			cpos, bpos: INTEGER
			positions: like index_to_positions
			byte: NATURAL_8
		do
			positions := index_to_positions (v)
			cpos := positions.item (1)
			bpos := positions.item (2)
			byte := map.item (cpos).code.to_natural_8
			byte := byte.set_bit (st, bpos)
			map.put (byte.to_character_8, cpos)
		ensure
			is_set: position_is_set (v) = st
		end

	--|--------------------------------------------------------------
invariant
	has_map: map /= Void
	consistent_size: count = map.count * 8

end -- class AEL_DS_BITMAP
